/*****************************************************************************
 *
 *      quickcamex.c - Logitech Quickcam Express USB kernel driver
 *
 *      Copyright (C) 2000 Georg Acher (acher@in.tum.de)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *      This driver is based on dabusb.c by Deti Fliegl (deti@fliegl.de)
 *
 *
 *****************************************************************************/

#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <linux/module.h>
#include <linux/socket.h>
#include <linux/miscdevice.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/kdev_t.h>

#include "quickcamex.h"
//#define dbg(format, arg...) printk(format "\n", ## arg)

/* --------------------------------------------------------------------- */

#define NRQUICKCAM 4

/*-------------------------------------------------------------------*/

static quickcam_t quickcam[NRQUICKCAM];
static int buffers = 256;

/*-------------------------------------------------------------------*/

static int quickcam_add_buf_tail (quickcam_t* s, struct list_head *dst, struct list_head *src)
{
	unsigned long flags;
	struct list_head *tmp;
	int ret = 0;

	spin_lock_irqsave (&s->lock, flags);

	if (list_empty (src)) {
		// no elements in source buffer
		ret = -1;
		goto err;
	}
	tmp = src->next;
	list_del (tmp);
	list_add_tail (tmp, dst);

  err:	spin_unlock_irqrestore (&s->lock, flags);
	return ret;
}
/*-------------------------------------------------------------------*/
#ifdef DEBUG 
static void dump_urb (urb_t* urb)
{
	dbg("urb                   :%p", urb);
	dbg("next                  :%p", urb->next);
	dbg("dev                   :%p", urb->dev);
	dbg("pipe                  :%08X", urb->pipe);
	dbg("status                :%d", urb->status);
	dbg("transfer_flags        :%08X", urb->transfer_flags);
	dbg("transfer_buffer       :%p", urb->transfer_buffer);
	dbg("transfer_buffer_length:%d", urb->transfer_buffer_length);
	dbg("actual_length         :%d", urb->actual_length);
	dbg("setup_packet          :%p", urb->setup_packet);
	dbg("start_frame           :%d", urb->start_frame);
	dbg("number_of_packets     :%d", urb->number_of_packets);
	dbg("interval              :%d", urb->interval);
	dbg("error_count           :%d", urb->error_count);
	dbg("context               :%p", urb->context);
	dbg("complete              :%p", urb->complete);
}
#endif
/*-------------------------------------------------------------------*/
static int quickcam_cancel_queue (quickcam_t* s, struct list_head *q)
{
	unsigned long flags;
	struct list_head *p;
	buff_t *b;

	dbg("quickcam_cancel_queue");

	spin_lock_irqsave (&s->lock, flags);

	for (p = q->next; p != q; p = p->next) {
		b = list_entry (p, buff_t, buff_list);

#ifdef DEBUG
		dump_urb(b->urb);
#endif
		usb_unlink_urb (b->urb);
	}
	spin_unlock_irqrestore (&s->lock, flags);
	return 0;
}
/*-------------------------------------------------------------------*/
static int quickcam_free_queue (struct list_head *q)
{
	struct list_head *tmp;
	struct list_head *p;
	buff_t *b;

	dbg("quickcam_free_queue");
	for (p = q->next; p != q;) {
		b = list_entry (p, buff_t, buff_list);

#ifdef DEBUG 
		dump_urb(b->urb);
#endif
		if (b->urb->transfer_buffer)
			kfree (b->urb->transfer_buffer);
		usb_free_urb(b->urb);
		tmp = p->next;
		list_del (p);
		kfree (b);
		p = tmp;
	}

	return 0;
}
/*-------------------------------------------------------------------*/
static int quickcam_free_buffers (quickcam_t* s)
{
	unsigned long flags;
	dbg("quickcam_free_buffers");

	spin_lock_irqsave(&s->lock, flags);

	quickcam_free_queue (&s->free_buff_list);
	quickcam_free_queue (&s->rec_buff_list);

	spin_unlock_irqrestore(&s->lock, flags);

	s->got_mem = 0;
	return 0;
}
/*-------------------------------------------------------------------*/
static void quickcam_iso_complete (urb_t* urb)
{
	buff_t *b = urb->context;
	quickcam_t* s = b->s;
	int i;
	int len;
	int dst = 0;
	void *buf = urb->transfer_buffer;
	struct usb_device *dev=s->usbdev;

	dbg("quickcam_iso_complete");

	// process if URB was not killed
	if (urb->status != -ENOENT) {
		unsigned int pipe = usb_rcvisocpipe (dev, _QUICKCAM_ISOPIPE);
		int pipesize = usb_maxpacket (dev, pipe, usb_pipeout (pipe));
		for (i = 0; i < urb->number_of_packets; i++)
			if (!urb->iso_frame_desc[i].status) {
				len = urb->iso_frame_desc[i].actual_length;
				if (len <= pipesize) {
#if 0
					*((char*)buf+dst)=0xff;
					*((char*)buf+dst+1)=0x00;
					*((char*)buf+dst+2)=0xaa;
					*((char*)buf+dst+3)=0x55;
#endif
					memcpy (buf + dst , buf + urb->iso_frame_desc[i].offset, len);
					dst += len;
				}
				else
					err("quickcam_iso_complete: invalid len %d", len);
			}
			else
				warn("quickcam_iso_complete: corrupted packet status: %d", urb->iso_frame_desc[i].status);
		//	if (dst != urb->actual_length)
		//	err("dst!=urb->actual_length:%d!=%d", dst, urb->actual_length);
	}

	if (atomic_dec_and_test (&s->pending_io) && !s->remove_pending && s->state != _stopped) {
		s->overruns++;
		err("overrun (%d)", s->overruns);
	}
	wake_up (&s->wait);
}
/*-------------------------------------------------------------------*/
static int quickcam_alloc_buffers (quickcam_t* s)
{
	int buffers = 0;
	buff_t *b;
	unsigned int pipe = usb_rcvisocpipe (s->usbdev, _QUICKCAM_ISOPIPE);
	int pipesize = usb_maxpacket (s->usbdev, pipe, usb_pipeout (pipe));
	int packets = _ISOPIPESIZE / pipesize;
	int transfer_buffer_length = packets * pipesize;
	int i;

	dbg("quickcam_alloc_buffers pipesize:%d packets:%d transfer_buffer_len:%d",
		 pipesize, packets, transfer_buffer_length);

	while (buffers < (s->total_buffer_size << 10)) {
		b = (buff_t*) kmalloc (sizeof (buff_t), GFP_KERNEL);
		if (!b) {
			err("kmalloc(sizeof(buff_t))==NULL");
			goto err;
		}
		memset (b, sizeof (buff_t), 0);
		b->s = s;
		b->urb = usb_alloc_urb(packets);
		if (!b->urb) {
			err("usb_alloc_urb == NULL");
			kfree (b);
			goto err;
		}

		b->urb->transfer_buffer = kmalloc (transfer_buffer_length, GFP_KERNEL);
		if (!b->urb->transfer_buffer) {
			kfree (b->urb);
			kfree (b);
			err("kmalloc(%d)==NULL", transfer_buffer_length);
			goto err;
		}

		b->urb->transfer_buffer_length = transfer_buffer_length;
		b->urb->number_of_packets = packets;
		b->urb->complete = quickcam_iso_complete;
		b->urb->context = b;
		b->urb->dev = s->usbdev;
		b->urb->pipe = pipe;
		b->urb->transfer_flags = USB_ISO_ASAP;

		for (i = 0; i < packets; i++) {
			b->urb->iso_frame_desc[i].offset = i * pipesize;
			b->urb->iso_frame_desc[i].length = pipesize;
		}

		buffers += transfer_buffer_length;
		list_add_tail (&b->buff_list, &s->free_buff_list);
	}
	s->got_mem = buffers;

	return 0;

	err:
	quickcam_free_buffers (s);
	return -ENOMEM;
}

/*-------------------------------------------------------------------*/
static loff_t quickcam_llseek (struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}
/*-------------------------------------------------------------------*/
static int quickcam_stop (quickcam_t* s)
{
	dbg("quickcam_stop");

	s->state = _stopped;
	quickcam_cancel_queue (s, &s->rec_buff_list);

	dbg("pending_io: %d", s->pending_io.counter);

	s->pending_io.counter = 0;
	return 0;
}
/*-------------------------------------------------------------------*/
static int quickcam_startrek (quickcam_t* s)
{
	if (!s->got_mem && s->state != _started) {

		dbg("quickcam_startrek");

		if (quickcam_alloc_buffers (s) < 0)
			return -ENOMEM;
		quickcam_stop (s);
		s->state = _started;
		s->readptr = 0;
	}

	if (!list_empty (&s->free_buff_list)) {
		buff_t *end;
		int ret;
		
	while (!quickcam_add_buf_tail (s, &s->rec_buff_list, &s->free_buff_list)) {

			dbg("submitting: end:%p s->rec_buff_list:%p", s->rec_buff_list.prev, &s->rec_buff_list);

			end = list_entry (s->rec_buff_list.prev, buff_t, buff_list);
			
			end->urb->dev=s->usbdev;
			ret = usb_submit_urb (end->urb);
			if (ret) {
				err("usb_submit_urb returned:%d", ret);
				if (quickcam_add_buf_tail (s, &s->free_buff_list, &s->rec_buff_list))
					err("startrek: quickcam_add_buf_tail failed");
				break;
			}
			else
				atomic_inc (&s->pending_io);
		}
		dbg("pending_io: %d",s->pending_io.counter);
	}

	return 0;
}
/*-------------------------------------------------------------------*/
static ssize_t quickcam_read (struct file *file, char *buf, size_t count, loff_t * ppos)
{
	quickcam_t* s = (quickcam_t*) file->private_data;
	unsigned long flags;
	unsigned ret = 0;
	int rem;
	int cnt;
	buff_t *b;
	urb_t* urb = NULL;

	dbg("quickcam_read");

	if (*ppos)
		return -ESPIPE;

	if (s->remove_pending)
		return -EIO;


	if (!s->usbdev)
		return -EIO;

	while (count > 0) {
		quickcam_startrek (s);

		spin_lock_irqsave (&s->lock, flags);

		if (list_empty (&s->rec_buff_list)) {

			spin_unlock_irqrestore(&s->lock, flags);

			err("error: rec_buf_list is empty");
			goto err;
		}
		
		b = list_entry (s->rec_buff_list.next, buff_t, buff_list);
		urb = b->urb;

		spin_unlock_irqrestore(&s->lock, flags);

		if (urb->status == -EINPROGRESS) {
			if (file->f_flags & O_NONBLOCK)		// return nonblocking
			 {
				if (!ret)
					ret = -EAGAIN;
				goto err;
			}

			interruptible_sleep_on (&s->wait);

			if (signal_pending (current)) {
				if (!ret)
					ret = -ERESTARTSYS;
				goto err;
			}

			spin_lock_irqsave (&s->lock, flags);

			if (list_empty (&s->rec_buff_list)) {
				spin_unlock_irqrestore(&s->lock, flags);
				err("error: still no buffer available.");
				goto err;
			}
			spin_unlock_irqrestore(&s->lock, flags);
			s->readptr = 0;
		}
		if (s->remove_pending) {
			ret = -EIO;
			goto err;
		}

		rem = urb->actual_length - s->readptr;		// set remaining bytes to copy

		if (count >= rem)
			cnt = rem;
		else
			cnt = count;

		dbg("copy_to_user:%p %p %d",buf, urb->transfer_buffer + s->readptr, cnt);

		if (copy_to_user (buf, urb->transfer_buffer + s->readptr, cnt)) {
			err("read: copy_to_user failed");
			if (!ret)
				ret = -EFAULT;
			goto err;
		}

		s->readptr += cnt;
		count -= cnt;
		buf += cnt;
		ret += cnt;

		if (s->readptr == urb->actual_length) {
			// finished, take next buffer
			if (quickcam_add_buf_tail (s, &s->free_buff_list, &s->rec_buff_list))
				err("read: quickcam_add_buf_tail failed");
			s->readptr = 0;
		}
	}
      err:			//up(&s->mutex);
	return ret;
}
/*-------------------------------------------------------------------*/
static int quickcam_open (struct inode *inode, struct file *file)
{
	int devnum = MINOR (inode->i_rdev);
	quickcam_t* s;

	if (devnum < QUICKCAM_MINOR || devnum >= (QUICKCAM_MINOR + NRQUICKCAM))
		return -EIO;

	s = &quickcam[devnum - QUICKCAM_MINOR];

	dbg("quickcam_open");
	down (&s->mutex);

	while (!s->usbdev || s->opened) {
		up (&s->mutex);

		if (file->f_flags & O_NONBLOCK) {
			return -EBUSY;
		}
		schedule_timeout (HZ / 2);

		if (signal_pending (current)) {
			return -EAGAIN;
		}
		down (&s->mutex);
	}
	if (usb_set_interface (s->usbdev, _QUICKCAM_IF, 1) < 0) {
	        up (&s->mutex);
		err("set_interface failed");
		return -EINVAL;
	}
	s->opened = 1;
	up (&s->mutex);

	file->f_pos = 0;
	file->private_data = s;

	return 0;
}
/*-------------------------------------------------------------------*/
static int quickcam_release (struct inode *inode, struct file *file)
{
	quickcam_t* s = (quickcam_t*) file->private_data;

	dbg("quickcam_release");

	down (&s->mutex);
	quickcam_stop (s);
	quickcam_free_buffers (s);
	up (&s->mutex);

	if (!s->remove_pending) {
		if (usb_set_interface (s->usbdev, _QUICKCAM_IF, 0) < 0)
			err("set_interface failed");
	}
	else
		wake_up (&s->remove_ok);

	s->opened = 0;
	return 0;
}
/*-------------------------------------------------------------------*/
static int quickcam_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	quickcam_t* s = (quickcam_t*) file->private_data;
	ctrl_transfer_t *ctrl;
	int ret = 0;

	dbg("quickcam_ioctl");

	if (s->remove_pending)
		return -EIO;

	down (&s->mutex);

	if (!s->usbdev) {
		up (&s->mutex);
		return -EIO;
	}

	switch (cmd) {

	case IOCTL_QUICKCAM_CTRLW:
		ctrl = (ctrl_transfer_t*) kmalloc (sizeof (ctrl_transfer_t), GFP_KERNEL);

		if (!ctrl) {
			ret = -ENOMEM;
			break;
		}

		if (copy_from_user (ctrl, (void *) arg, sizeof (ctrl_transfer_t))) {
			ret = -EFAULT;
			kfree (ctrl);
			break;
		}

		ret=usb_control_msg(s->usbdev, usb_sndctrlpipe( s->usbdev, 0 ), 
				    ctrl->setup.request, ctrl->setup.requesttype,
				    ctrl->setup.value,ctrl->setup.index,				    
				    ctrl->data, ctrl->size,10);

		kfree (ctrl);
		break;
		
	case IOCTL_QUICKCAM_CTRLR:
		ctrl = (ctrl_transfer_t*) kmalloc (sizeof (ctrl_transfer_t), GFP_KERNEL);

		if (!ctrl) {
			ret = -ENOMEM;
			break;
		}

		if (copy_from_user (ctrl, (void *) arg, sizeof (ctrl_transfer_t))) {
			ret = -EFAULT;
			kfree (ctrl);
			break;
		}	

		ret=usb_control_msg(s->usbdev, usb_rcvctrlpipe( s->usbdev, 0 ), 
				    ctrl->setup.request, ctrl->setup.requesttype,
				    ctrl->setup.value,ctrl->setup.index,				    
				    ctrl->data, ctrl->size,10);

		if(ret>=0)
			ret = copy_to_user ((void *) arg, ctrl, sizeof(ctrl_transfer_t));

		kfree (ctrl);
		break;
		
	case IOCTL_QUICKCAM_ID:
		ret=s->usbdev->descriptor.idProduct;
		break;

	default:
		ret = -ENOIOCTLCMD;
		break;
	}
	up (&s->mutex);
	return ret;
}

static struct file_operations quickcam_fops =
{
// when did owner occur?
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,10)
	owner:		THIS_MODULE,
#endif
	llseek:		quickcam_llseek,
	read:		quickcam_read,
	ioctl:		quickcam_ioctl,
	open:		quickcam_open,
	release:	quickcam_release,
};

static int quickcam_find_struct (void)
{
	int u;

	for (u = 0; u < NRQUICKCAM; u++) {
		quickcam_t* s = &quickcam[u];
		if (!s->usbdev)
			return u;
	}
	return -1;
}

/* --------------------------------------------------------------------- */
static void *quickcam_probe (struct usb_device *usbdev, unsigned ifnum,
			     const struct usb_device_id* id)
{
	int devnum;
	quickcam_t* s;

	printk("quickcam: probe: vendor id 0x%x, device id 0x%x ifnum:%d",
	  usbdev->descriptor.idVendor, usbdev->descriptor.idProduct, ifnum);

	if (usbdev->descriptor.idVendor != 0x046d)
		if  ((usbdev->descriptor.idProduct != 0x0840) && // Quickcam Express
		     (usbdev->descriptor.idProduct != 0x0850)) // Lego Cam
		return NULL;

	/* We don't handle multiple configurations */
	if (usbdev->descriptor.bNumConfigurations != 1)
		return NULL;

	devnum = quickcam_find_struct ();
	if (devnum == -1)
		return NULL;

	s = &quickcam[devnum];

	down (&s->mutex);
	s->remove_pending = 0;
	s->usbdev = usbdev;

	if (usb_set_configuration (usbdev, usbdev->config[0].bConfigurationValue) < 0) {
		err("set_configuration failed");
		goto reject;
	}
	else {
		if (usb_set_interface (s->usbdev, _QUICKCAM_IF, 0) < 0) {
			err("set_interface failed");
			goto reject;
		}
	}
	printk("bound to interface: %d\n", devnum);
	up (&s->mutex);
	//	MOD_INC_USE_COUNT;
	return s;

      reject:
	up (&s->mutex);
	s->usbdev = NULL;
	return NULL;
}

static void quickcam_disconnect (struct usb_device *usbdev, void *ptr)
{
	quickcam_t* s = (quickcam_t*) ptr;

	dbg("quickcam_disconnect");

	s->remove_pending = 1;
	wake_up (&s->wait);
	if (s->state == _started)
		sleep_on (&s->remove_ok);
	s->usbdev = NULL;
	s->overruns = 0;
	//	MOD_DEC_USE_COUNT;
}

static struct usb_driver quickcam_driver =
{
	"quickcam",
	quickcam_probe,
	quickcam_disconnect,
	{NULL, NULL},
	&quickcam_fops,
	QUICKCAM_MINOR
};

/* --------------------------------------------------------------------- */

int __init quickcam_init (void)
{
	unsigned u;

	/* initialize struct */
	for (u = 0; u < NRQUICKCAM; u++) {
		quickcam_t* s = &quickcam[u];
		memset (s, 0, sizeof (quickcam_t));
		init_MUTEX (&s->mutex);
		s->usbdev = NULL;
		s->total_buffer_size = buffers;
		init_waitqueue_head (&s->wait);
		init_waitqueue_head (&s->remove_ok);
		spin_lock_init (&s->lock);
		INIT_LIST_HEAD (&s->free_buff_list);
		INIT_LIST_HEAD (&s->rec_buff_list);
	}

	/* register misc device */
	if (usb_register(&quickcam_driver))
		return -1;

	dbg("quickcam_init: driver registered");
	return 0;
}

void __exit quickcam_cleanup (void)
{
	dbg("quickcam_cleanup");

	usb_deregister (&quickcam_driver);
}

/* --------------------------------------------------------------------- */

#ifdef MODULE
MODULE_AUTHOR ("Georg Acher, acher@in.tum.de");
MODULE_DESCRIPTION ("Logitech Quickcam Express USB driver for Linux (c)2000");

int __init init_module (void)
{
	return quickcam_init ();
}

void __exit cleanup_module (void)
{
	quickcam_cleanup ();
}

#endif

/* --------------------------------------------------------------------- */
